from __future__ import annotations

import os
import stat
from functools import lru_cache
from pathlib import Path
from typing import Generator, Optional

from tach.colors import BCOLORS
from tach.utils.exclude import is_path_excluded


def write_file(path: Path, content: str, root: Optional[Path] = None):
    path.write_text(content)
    try:
        display_path = path.relative_to(root or Path.cwd())
    except AssertionError:
        display_path = path.resolve()
    print(f"{BCOLORS.WARNING}Wrote '{display_path}'{BCOLORS.ENDC}")


def mark_executable(path: Path):
    path.chmod(path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)


def walk(
    root: Path,
    project_root: Optional[Path] = None,
    exclude_paths: Optional[list[str]] = None,
    use_regex_matching: bool = False,
    depth: int | None = None,
) -> Generator[tuple[Path, list[Path]], None, None]:
    if depth is not None and depth <= 0:
        return
    root = root.resolve()
    for dirpath, dirnames, filenames in os.walk(root):
        rel_dirpath = Path(dirpath).relative_to(root)

        if rel_dirpath.name.startswith("."):
            # This prevents recursing into child directories of hidden paths
            del dirnames[:]
            continue

        if exclude_paths:
            project_dirpath = (
                Path(dirpath).relative_to(project_root) if project_root else rel_dirpath
            )
            if is_path_excluded(
                exclude_paths, project_dirpath, use_regex_matching=use_regex_matching
            ):
                del dirnames[:]
                continue

        if depth:
            # Ignore anything past requested depth
            current_depth = len(rel_dirpath.parts) - 1
            if current_depth > depth:
                continue

        def filter_filename(filename: str) -> bool:
            return not filename.startswith(".")

        yield rel_dirpath, list(map(Path, filter(filter_filename, filenames)))


def walk_pyfiles(
    root: Path,
    project_root: Optional[Path] = None,
    exclude_paths: Optional[list[str]] = None,
    use_regex_matching: bool = False,
    depth: int | None = None,
) -> Generator[Path, None, None]:
    for dirpath, filepaths in walk(
        root,
        project_root=project_root,
        exclude_paths=exclude_paths,
        use_regex_matching=use_regex_matching,
        depth=depth,
    ):
        for filepath in filepaths:
            if filepath.name.endswith(".py"):
                yield dirpath / filepath


@lru_cache(maxsize=None)
def file_to_module_path(source_roots: tuple[Path, ...], file_path: Path) -> str:
    # NOTE: source_roots are assumed to be absolute here
    matching_root: Path | None = None
    for root in source_roots:
        if root in file_path.parents:
            matching_root = root
            break

    if matching_root is None:
        raise ValueError(f"File path: {file_path} not found in any source root.")

    relative_path = file_path.relative_to(matching_root)
    components = list(relative_path.parent.parts)

    if relative_path.name != "__init__.py":
        components.append(relative_path.stem)

    module_path = ".".join(components)
    return "." if not module_path else module_path


@lru_cache(maxsize=None)
def module_to_pyfile_or_dir_path(
    source_roots: tuple[Path, ...], module_path: str
) -> Path | None:
    """
    This resolves a dotted Python module path ('a.b.c')
    into a Python file or a Python package directory,
    used in cases where __init__.py is not relevant such as the
    interactive module tree.

    The module path is assumed NOT to refer to a member within a Python module.

    'source_roots' is assumed to be a list of absolute paths.
    """
    if not module_path:
        # Path("") turns into PosixPath("."), but we don't want to
        # treat an empty module path as the root directory
        return None

    base_path = module_path.replace(".", os.sep)
    for source_root in source_roots:
        dir_path = source_root / base_path
        pyinterface_path = source_root / f"{base_path}.pyi"
        pyfile_path = source_root / f"{base_path}.py"
        if dir_path.is_dir():
            return dir_path
        elif pyinterface_path.exists():
            return pyinterface_path
        elif pyfile_path.exists():
            return pyfile_path

    return None



from __future__ import annotations

import os
import subprocess
import sys
import threading
from dataclasses import dataclass
from queue import Queue
from typing import IO, TYPE_CHECKING, Any, Tuple

from tach.errors import TachSetupError

if TYPE_CHECKING:
    from pathlib import Path

    from tach.extension import ProjectConfig


def run_and_capture(cmd: list[str], **kwargs: Any) -> Tuple[int, str, str]:
    stdout_queue: Queue[str] = Queue()
    stderr_queue: Queue[str] = Queue()

    def tee_output(pipe: IO[str], queue: Queue[str], terminal: Any):
        for line in pipe:
            queue.put_nowait(line)
            print(line, end="", file=terminal, flush=True)
        pipe.close()

    process_env = os.environ.copy()
    process_env["PYTEST_ADDOPTS"] = "--color=yes"
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
        env=process_env,
        **kwargs,
    )

    stdout_thread = threading.Thread(
        target=tee_output, args=(process.stdout, stdout_queue, sys.stdout)
    )
    stderr_thread = threading.Thread(
        target=tee_output, args=(process.stderr, stderr_queue, sys.stderr)
    )

    stdout_thread.start()
    stderr_thread.start()

    returncode = process.wait()
    stdout_thread.join()
    stderr_thread.join()

    stdout_content = "".join(stdout_queue.queue)
    stderr_content = "".join(stderr_queue.queue)

    return returncode, stdout_content, stderr_content


@dataclass
class AffectedTestsResult:
    exit_code: int
    tests_ran_to_completion: bool
    stdout: str
    stderr: str


def run_affected_tests(
    project_root: Path,
    project_config: ProjectConfig,
    head: str,
    base: str,
    pytest_args: list[Any] | None = None,
) -> AffectedTestsResult:
    try:
        import pytest  # type: ignore  # noqa: F401
    except ImportError:
        raise TachSetupError("Cannot run tests, could not find 'pytest'.")

    cmd = ["pytest", "-p", "tach.pytest_plugin"]
    if pytest_args:
        cmd.extend(pytest_args)
    if base:
        cmd.extend(["--tach-base", base])
    if head:
        cmd.extend(["--tach-head", head])

    returncode, stdout, stderr = run_and_capture(cmd, cwd=project_root)
    tests_ran = returncode != pytest.ExitCode.NO_TESTS_COLLECTED
    exit_code = (
        pytest.ExitCode.OK
        if returncode == pytest.ExitCode.NO_TESTS_COLLECTED
        else returncode
    )

    return AffectedTestsResult(
        exit_code=exit_code,
        tests_ran_to_completion=tests_ran,
        stdout=stdout,
        stderr=stderr,
    )


__all__ = ["run_affected_tests"]



from __future__ import annotations

from typing import TYPE_CHECKING

import tomli
import tomli_w

from tach import errors
from tach import filesystem as fs
from tach.extension import (
    sync_project as sync_project_ext,
)
from tach.filesystem import get_project_config_path

if TYPE_CHECKING:
    from pathlib import Path

    from tach.extension import ProjectConfig


def sync_project(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
    add: bool = False,
) -> None:
    config_path = get_project_config_path(project_root)
    if config_path is None:
        raise errors.TachError(
            "Unexpected error. Could not find configuration file during 'sync'."
        )

    config_toml_content = sync_project_ext(
        project_root=project_root,
        project_config=project_config,
        exclude_paths=exclude_paths,
        add=add,
    )
    # Format the content, TODO: should'nt be handled here
    config_toml_content = tomli_w.dumps(tomli.loads(config_toml_content))
    fs.write_file(config_path, config_toml_content, root=project_root)


__all__ = ["sync_project"]



# type: ignore reportConstantRedefinition
from __future__ import annotations

import sys


def test_unicode_encoding():
    try:
        "✅".encode(sys.stdout.encoding)
        return True
    except UnicodeEncodeError:
        return False


SUPPORTS_UNICODE = test_unicode_encoding()


### Icons which don't depend on Unicode support
WARNING = "‼️"


### Icons which depend on Unicode support
if SUPPORTS_UNICODE:
    SUCCESS = "✅"
    FAIL = "❌"
else:
    SUCCESS = "[OK]"
    FAIL = "[FAIL]"



from pathlib import Path

from setuptools import setup


# For this trick, I thank:
# https://stackoverflow.com/a/78056725

root: Path = Path(__file__).parent

if __name__ == "__main__":
    setup(
        install_requires=[
            f"myorg-pack-{letter} @ {(root.parent / f'pack-{letter}').as_uri()}"
            for letter in ["a", "b", "c", "d", "e", "f"]
        ],
    )



from __future__ import annotations

import shutil
import tempfile
from pathlib import Path

import pytest

from tach.cli import tach_sync


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_project_root = Path(temp_dir) / "valid"
        shutil.copytree(project_root, temp_project_root)

        with pytest.raises(SystemExit) as exc_info:
            tach_sync(project_root=temp_project_root, add=True)

        assert exc_info.value.code == 0
        captured = capfd.readouterr()
        assert "✅" in captured.out  # success state



from __future__ import annotations

import json
from http.client import HTTPSConnection
from typing import TYPE_CHECKING, Any
from urllib import parse

if TYPE_CHECKING:
    import uuid

LOGGING_URL = "https://vmilasesnyvpalekembc.supabase.co"
PUBLIC_ANON_CLIENT_KEY = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6InZtaWxhc2Vzbnl2cGFsZWtlbWJjIiwicm9"
    "sZSI6ImFub24iLCJpYXQiOjE3MTY0OTEwOTMsImV4cCI6MjAzMjA2NzA5M30.ndk9sUAmMJ5oNenDmLw35uT0s_d6c56Hk_PL5BucrOc"
)


def log_request(url: str, data: dict[str, Any]) -> None:
    headers = {
        "Content-Type": "application/json",
        "apikey": PUBLIC_ANON_CLIENT_KEY,
        "authorization": f"Bearer {PUBLIC_ANON_CLIENT_KEY}",
    }
    json_data = json.dumps(data)
    full_url = f"{LOGGING_URL}/{url}"
    conn = None
    try:
        url_parts: parse.ParseResult = parse.urlparse(full_url)
        conn = HTTPSConnection(url_parts.netloc, timeout=1)
        conn.request("POST", full_url, body=json_data, headers=headers)
        conn.getresponse()
    except Exception:  # noqa
        pass
    finally:
        if conn is not None:
            conn.close()


def log_uid(uid: uuid.UUID, is_ci: bool, is_gauge: bool) -> None:
    log_request(
        url="rest/v1/User", data={"id": str(uid), "is_ci": is_ci, "is_gauge": is_gauge}
    )


def log_record(record_data: dict[str, Any]) -> None:
    log_request(url="rest/v1/LogRecord", data=record_data)
